# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for pyglove.symbolic.base."""

import copy
import dataclasses
import unittest

from pyglove.core import object_utils
from pyglove.core import typing as pg_typing
from pyglove.core.symbolic import base
from pyglove.core.symbolic.dict import Dict


class FieldUpdateTest(unittest.TestCase):
  """Tests for `pg.symbolic.FieldUpdate`."""

  def test_basics(self):
    x = Dict(x=1)
    f = pg_typing.Field('x', pg_typing.Int())
    update = base.FieldUpdate(object_utils.KeyPath('x'), x, f, 1, 2)
    self.assertEqual(update.path, 'x')
    self.assertIs(update.target, x)
    self.assertIs(update.field, f)
    self.assertEqual(update.old_value, 1)
    self.assertEqual(update.new_value, 2)

  def test_format(self):
    self.assertEqual(
        base.FieldUpdate(
            object_utils.KeyPath('x'), Dict(x=1), None, 1, 2
        ).format(compact=True),
        'FieldUpdate(parent_path=, path=x, old_value=1, new_value=2)',
    )

    self.assertEqual(
        base.FieldUpdate(
            object_utils.KeyPath('a'), Dict(x=Dict(a=1)).x, None, 1, 2
        ).format(compact=True),
        'FieldUpdate(parent_path=x, path=a, old_value=1, new_value=2)',
    )

  def test_eq_ne(self):
    x = Dict()
    f = pg_typing.Field('x', pg_typing.Int())
    self.assertEqual(
        base.FieldUpdate(object_utils.KeyPath('a'), x, f, 1, 2),
        base.FieldUpdate(object_utils.KeyPath('a'), x, f, 1, 2),
    )

    # Targets are not the same instance.
    self.assertNotEqual(
        base.FieldUpdate(object_utils.KeyPath('a'), x, f, 1, 2),
        base.FieldUpdate(object_utils.KeyPath('a'), Dict(), f, 1, 2),
    )

    # Fields are not the same instance.
    self.assertNotEqual(
        base.FieldUpdate(object_utils.KeyPath('a'), x, f, 1, 2),
        base.FieldUpdate(object_utils.KeyPath('b'), x, copy.copy(f), 1, 2),
    )

    self.assertNotEqual(
        base.FieldUpdate(object_utils.KeyPath('a'), x, f, 1, 2),
        base.FieldUpdate(object_utils.KeyPath('a'), x, f, 0, 2),
    )

    self.assertNotEqual(
        base.FieldUpdate(object_utils.KeyPath('a'), x, f, 1, 2),
        base.FieldUpdate(object_utils.KeyPath('a'), x, f, 1, 1),
    )

    self.assertNotEqual(
        base.FieldUpdate(object_utils.KeyPath('a'), x, f, 1, 2), Dict()
    )


class ContextualValueTest(unittest.TestCase):
  """Tests for `pg.symbolic.ContextualValue`."""

  def test_str(self):
    self.assertEqual(str(base.ContextualValue()), 'ContextualValue()')

  def test_repr(self):
    self.assertEqual(repr(base.ContextualValue()), 'ContextualValue()')

  def test_eq(self):
    self.assertEqual(base.ContextualValue(), base.ContextualValue())
    self.assertNotEqual(base.ContextualValue(), 1)

  def test_call(self):
    @dataclasses.dataclass
    class A:
      x: int = 1
      y: int = 2

    self.assertEqual(
        base.ContextualValue().get(base.GetAttributeContext('x', A(), Dict())),
        1,
    )
    self.assertEqual(
        base.ContextualValue().get(base.GetAttributeContext(0, [0, 1], Dict())),
        0,
    )
    self.assertEqual(
        base.ContextualValue().get(base.GetAttributeContext(0, Dict(), Dict())),
        pg_typing.MISSING_VALUE,
    )

  def test_custom_typing(self):
    v = base.ContextualValue()
    self.assertIs(pg_typing.Int().apply(v), v)
    self.assertIs(pg_typing.Str().apply(v), v)

  def test_to_json(self):
    self.assertEqual(
        base.to_json(base.ContextualValue()),
        {'_type': f'{base.ContextualValue.__module__}.ContextualValue'},
    )

  def test_from_json(self):
    self.assertEqual(
        base.from_json(base.ContextualValue().to_json()), base.ContextualValue()
    )


if __name__ == '__main__':
  unittest.main()



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common traits for Python objects.

This file defines interfaces for describing the common traits of a Python
object, for example, serialization (JSONConvertible), partiality (MaybePartial),
formatting (Formattable), functor (Functor).
"""

import abc
import inspect
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Type, TypeVar, Union

# Nestable[T] is a (maybe) nested structure of T, which could be T, a Dict
# a List or a Tuple of Nestable[T]. We use a Union to fool PyType checker to
# make Nestable[T] a valid type annotation without type check.
T = TypeVar('T')
Nestable = Union[Any, T]  # pytype: disable=not-supported-yet

# pylint: disable=invalid-name
JSONPrimitiveType = Union[int, float, bool, str]

# pytype doesn't support recursion. Use Any instead of 'JSONValueType'
# in List and Dict.
JSONListType = List[Any]
JSONDictType = Dict[str, Any]
JSONValueType = Union[JSONPrimitiveType, JSONListType, JSONDictType]

# pylint: enable=invalid-name


class _TypeRegistry:
  """A registry for mapping a string name to type definition.

  This class is used for looking up type definition by a string identifier for
  deserialization.
  """

  def __init__(self):
    """Constructor."""
    # NOTE(daiyip): the order of keys in the dict is preserved. As a result,
    # in `pg.wrapping.apply_wrappers`, the latest registered wrapper
    # class will always be picked up when there are multiple wrapper classes
    # registered for a user class.
    self._type_to_cls_map = dict()

  def register(
      self, type_name: str, cls: Type[Any], override_existing: bool = False
      ) -> None:
    """Register a ``symbolic.Object`` class with a type name.

    Args:
      type_name: String identifier for the class, which will be used as the
        value of `_type` property when deciding which class to construct object
        when converting a JSON value to object.
      cls: Class to register.
      override_existing: Whether allow to override existing value if type name
        is already registered.

    Raises:
      KeyError: If type_name is already registered and override_existing is set
        to False.
    """
    if type_name in self._type_to_cls_map and not override_existing:
      raise KeyError(
          f'Type {type_name!r} has already been registered with class '
          f'{self._type_to_cls_map[type_name].__name__}.')
    self._type_to_cls_map[type_name] = cls

  def is_registered(self, type_name: str) -> bool:
    """Returns whether a type name is registered."""
    return type_name in self._type_to_cls_map

  def class_from_typename(
      self, type_name: str) -> Optional[Type[Any]]:
    """Get class from type name."""
    return self._type_to_cls_map.get(type_name, None)

  def iteritems(self) -> Iterable[Tuple[str, Type[Any]]]:
    """Iterate type registry."""
    return self._type_to_cls_map.items()


class JSONConvertible(metaclass=abc.ABCMeta):
  """Interface for classes whose instances are convertible from/to JSON.

  A JSON convertible object is an object that can be converted into plain Python
  objects, hence can be serialized into or deserialized from JSON.

  Subclasses of ``JSONConvertible`` should implement:

    * ``to_json``: A method that returns a plain Python dict with a `_type`
      property whose value should identify the class.
    * ``from_json``: A class method that takes a plain Python dict and returns
      an instance of the class.

  Example::

    class MyObject(pg.JSONConvertible):

      def __init__(self, x: int):
        self.x = x

      def to_json(self, **kwargs):
        return {
          '_type': 'MyObject',
          'x': self.x
        }

      @classmethod
      def from_json(cls, json_value, **kwargs):
        return cls(json_value['x'])

  All symbolic types (see :class:`pyglove.Symbolic`) are JSON convertible.
  """

  # Registry for looking up the type definition for a string identifier during
  # deserialization. One key can be used for only one type, while the same type
  # can be registered with many different string identifiers, which can be
  # useful to allow backward compatibility of existing serialized strings.
  _TYPE_REGISTRY = _TypeRegistry()

  # Key in serialized dict that represents the class to restore.
  TYPE_NAME_KEY = '_type'

  # Marker (as the first element of a list) for serializing tuples.
  TUPLE_MARKER = '__tuple__'

  # Type converter that converts a complex type to basic JSON value type.
  # When this field is set by users, the converter will be invoked when a
  # complex value cannot be serialized by existing methods.
  TYPE_CONVERTER: Optional[
      Callable[[Type[Any]], Callable[[Any], JSONValueType]]] = None

  # Class property that indicates whether to automatically register class
  # for deserialization. Subclass can override.
  auto_register = True

  @classmethod
  def from_json(cls, json_value: JSONValueType, **kwargs) -> 'JSONConvertible':
    """Creates an instance of this class from a plain Python value.

    NOTE(daiyip): ``pg.Symbolic`` overrides ``from_json`` class method.

    Args:
      json_value: JSON value type.
      **kwargs: Keyword arguments as flags to control object creation.

    Returns:
      An instance of cls.
    """
    del kwargs
    assert isinstance(json_value, dict)
    init_args = {k: from_json(v) for k, v in json_value.items()
                 if k != JSONConvertible.TYPE_NAME_KEY}
    return cls(**init_args)

  @abc.abstractmethod
  def to_json(self, **kwargs) -> JSONValueType:
    """Returns a plain Python value as a representation for this object.

    A plain Python value are basic python types that can be serialized into
    JSON, e.g: ``bool``, ``int``, ``float``, ``str``, ``dict`` (with string
    keys), ``list``, ``tuple`` where the container types should have plain
    Python values as their values.

    Args:
      **kwargs: Keyword arguments as flags to control JSON conversion.

    Returns:
      A plain Python value.
    """

  @classmethod
  def register(
      cls,
      type_name: str,
      subclass: Type['JSONConvertible'],
      override_existing: bool = False
      ) -> None:
    """Registers a class with a type name.

    The type name will be used as the key for class lookup during
    deserialization. A class can be registered with multiple type names, but
    a type name should be uesd only for one class.

    Args:
      type_name: A global unique string identifier for subclass.
      subclass: A subclass of JSONConvertible.
      override_existing: If True, registering an type name is allowed.
        Otherwise an error will be raised.
    """
    cls._TYPE_REGISTRY.register(type_name, subclass, override_existing)

  @classmethod
  def is_registered(cls, type_name: str) -> bool:
    """Returns True if a type name is registered. Otherwise False."""
    return cls._TYPE_REGISTRY.is_registered(type_name)

  @classmethod
  def class_from_typename(
      cls, type_name: str) -> Optional[Type['JSONConvertible']]:
    """Gets the class for a registered type name.

    Args:
      type_name: A string as the global unique type identifier for requested
        class.

    Returns:
      A type object if registered, otherwise None.
    """
    return cls._TYPE_REGISTRY.class_from_typename(type_name)

  @classmethod
  def registered_types(cls) -> Iterable[Tuple[str, Type['JSONConvertible']]]:
    """Returns an iterator of registered (serialization key, class) tuples."""
    return cls._TYPE_REGISTRY.iteritems()

  @classmethod
  def to_json_dict(
      cls,
      fields: Dict[str, Any],
      **kwargs) -> Dict[str, JSONValueType]:
    """Helper method to create JSON dict from class and field."""
    type_name = f'{cls.__module__}.{cls.__name__}'
    json_dict = {JSONConvertible.TYPE_NAME_KEY: type_name}
    json_dict.update({k: to_json(v, **kwargs) for k, v in fields.items()})
    return json_dict

  @classmethod
  def __init_subclass__(cls):
    super().__init_subclass__()
    if not inspect.isabstract(cls) and cls.auto_register:
      type_name = f'{cls.__module__}.{cls.__name__}'
      JSONConvertible.register(type_name, cls, override_existing=True)


def registered_types() -> Iterable[Tuple[str, Type[JSONConvertible]]]:
  """Returns an iterator of registered (serialization key, class) tuples."""
  return JSONConvertible.registered_types()


def to_json(value: Any, **kwargs) -> Any:
  """Serializes a (maybe) JSONConvertible value into a plain Python object.

  Args:
    value: value to serialize. Applicable value types are:

      * Builtin python types: None, bool, int, float, string;
      * JSONConvertible types;
      * List types;
      * Tuple types;
      * Dict types.

    **kwargs: Keyword arguments to pass to value.to_json if value is
      JSONConvertible.

  Returns:
    JSON value.
  """
  if isinstance(value, (type(None), bool, int, float, str)):
    return value
  elif isinstance(value, JSONConvertible):
    return value.to_json(**kwargs)
  elif isinstance(value, tuple):
    return [JSONConvertible.TUPLE_MARKER] + to_json(list(value), **kwargs)
  elif isinstance(value, list):
    return [to_json(item, **kwargs) for item in value]
  elif isinstance(value, dict):
    return {k: to_json(v, **kwargs) for k, v in value.items()}
  else:
    if JSONConvertible.TYPE_CONVERTER is not None:
      converter = JSONConvertible.TYPE_CONVERTER(type(value))   # pylint: disable=not-callable
      if converter:
        return to_json(converter(value))
    raise ValueError(f'Cannot convert complex type {value} to JSON.')


def from_json(json_value: JSONValueType) -> Any:
  """Deserializes a (maybe) JSONConvertible value from JSON value.

  Args:
    json_value: Input JSON value.

  Returns:
    Deserialized value.
  """
  if isinstance(json_value, list):
    if json_value and json_value[0] == JSONConvertible.TUPLE_MARKER:
      if len(json_value) < 2:
        raise ValueError(
            f'Tuple should have at least one element '
            f'besides \'{JSONConvertible.TUPLE_MARKER}\'. '
            f'Encountered: {json_value}.')
      return tuple([from_json(v) for v in json_value[1:]])
    return [from_json(v) for v in json_value]
  elif isinstance(json_value, dict):
    if JSONConvertible.TYPE_NAME_KEY not in json_value:
      return {k: from_json(v) for k, v in json_value.items()}
    type_name = json_value[JSONConvertible.TYPE_NAME_KEY]
    cls = JSONConvertible.class_from_typename(type_name)
    if cls is None:
      raise TypeError(
          f'Type name \'{type_name}\' is not registered '
          f'with a `pg.JSONConvertible` subclass.')
    return cls.from_json(json_value)
  return json_value


class Formattable(metaclass=abc.ABCMeta):
  """Interface for classes whose instances can be pretty-formatted.

  This interface overrides the default ``__repr__`` and ``__str__`` method, thus
  all ``Formattable`` objects can be printed nicely.

  All symbolic types implement this interface.
  """

  @abc.abstractmethod
  def format(self,
             compact: bool = False,
             verbose: bool = True,
             root_indent: int = 0,
             **kwargs) -> str:
    """Formats this object into a string representation.

    Args:
      compact: If True, this object will be formatted into a single line.
      verbose: If True, this object will be formatted with verbosity.
        Subclasses should define `verbosity` on their own.
      root_indent: The start indent level for this object if the output is a
        multi-line string.
      **kwargs: Subclass specific keyword arguments.

    Returns:
      A string of formatted object.
    """

  def __str__(self) -> str:
    """Returns the full (maybe multi-line) representation of this object."""
    return self.format(compact=False, verbose=True)

  def __repr__(self) -> str:
    """Returns a single-line representation of this object."""
    return self.format(compact=True)


class MaybePartial(metaclass=abc.ABCMeta):
  """Interface for classes whose instances can be partially constructed.

  A ``MaybePartial`` object is an object whose ``__init__`` method can accept
  ``pg.MISSING_VALUE`` as its argument values. All symbolic types (see
  :class:`pyglove.Symbolic`) implements this interface, as their symbolic
  attributes can be partially filled.

  Example::

    d = pg.Dict(x=pg.MISSING_VALUE, y=1)
    assert d.is_partial
    assert 'x' in d.missing_values()
  """

  @property
  def is_partial(self) -> bool:
    """Returns True if this object is partial. Otherwise False.

    An object is considered partial when any of its required fields is missing,
    or at least one member is partial. The subclass can override this method
    to provide a more efficient solution.
    """
    return len(self.missing_values()) > 0  # pylint: disable=g-explicit-length-test

  @abc.abstractmethod
  def missing_values(self, flatten: bool = True) -> Dict[str, Any]:  # pylint: disable=redefined-outer-name
    """Returns missing values from this object.

    Args:
      flatten: If True, convert nested structures into a flattened dict using
        key path (delimited by '.' and '[]') as key.

    Returns:
      A dict of key to MISSING_VALUE.
    """


class Functor(metaclass=abc.ABCMeta):
  """Interface for functor."""

  @abc.abstractmethod
  def __call__(self, *args, **kwargs) -> Any:
    """Calls the functor.

    Args:
      *args: Any positional arguments.
      **kwargs: Any keyword arguments.

    Returns:
      Any value.
    """



# Copyright 2023 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Customizable contextual value markers."""
from typing import Any
from pyglove.core.symbolic import base
from pyglove.core.symbolic import functor


class ContextualGetter(functor.Functor, base.ContextualValue):
  """Base for functor-based contextual getter."""

  def value_from(self, context: base.GetAttributeContext) -> Any:
    return self.__call__(context)


def contextual_getter(args=None, returns=None, **kwargs):
  """Decorator that makes ContextualGetter class from function.

  Examples::

    @pg.contextual_getter
    def static_value(self, name, context, value):
      return value

    class A(pg.Object):
      x: pg.ContextualValue = static_value(value=1)

  Args:
    args: A list of tuples that defines the schema for function arguments.
      Please see `functor_class` for detailed explanation of `args`. If None, it
      will be inferenced from the function argument annotations.
    returns: Optional value spec for return value. If None, it will be inferred
      from the function return value annotation.
    **kwargs: Additional keyword argments for controlling the behavior of
      functor creation. Please refer to :func:`pg.symbolic.functor_class` for
      more details.

  Returns:
    A function that converts a regular function into a ``pg.ContextualGetter``
      subclass.
  """
  return functor.functor(args, returns, base_class=ContextualGetter, **kwargs)



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Automatic type conversions."""

import calendar
import datetime
from typing import Any, Callable, Optional, Tuple, Type, Union

from pyglove.core import object_utils


class _TypeConverterRegistry:
  """Type converter registry."""

  def __init__(self):
    """Constructor."""
    self._converter_list = []
    self._json_value_types = set(
        [int, float, bool, type(None), list, tuple, dict, str])

  def register(
      self,
      src: Union[Type[Any], Tuple[Type[Any], ...]],
      dest: Union[Type[Any], Tuple[Type[Any], ...]],
      convert_fn: Callable[[Any], Any]) -> None:  # pyformat: disable pylint: disable=line-too-long
    """Register a converter from src type to dest type."""
    if (not isinstance(src, (tuple, type)) or
        not isinstance(dest, (tuple, type))):
      raise TypeError('Argument \'src\' and \'dest\' must be a type or '
                      'tuple of types.')
    if isinstance(dest, tuple):
      json_value_convertible = any(d in self._json_value_types for d in dest)
    else:
      json_value_convertible = dest in self._json_value_types
    self._converter_list.append((src, dest, convert_fn, json_value_convertible))

  def get_converter(
      self, src: Type[Any], dest: Type[Any]) -> Optional[Callable[[Any], Any]]:
    """Get converter from source type to destination type."""
    # TODO(daiyip): Right now we don't see the need of a large number of
    # converters, thus its affordable to iterate the list.
    # We may consider more efficient way to do lookup in future.
    # NOTE(daiyip): We do reverse lookup since usually subclass converter
    # is register after base class.
    for src_type, dest_type, converter, _ in reversed(self._converter_list):
      dest_type = dest_type if isinstance(dest_type, tuple) else (dest_type,)
      if issubclass(src, src_type) and dest in dest_type:
        return converter
    return None

  def get_json_value_converter(
      self, src: Type[Any]) -> Optional[Callable[[Any], Any]]:
    """Get converter from source type to a JSON simple type."""
    for src_type, _, converter, json_value_convertible in reversed(
        self._converter_list):
      if issubclass(src, src_type) and json_value_convertible:
        return converter
    return None


_TYPE_CONVERTER_REGISTRY = _TypeConverterRegistry()


def get_converter(
    src: Type[Any], dest: Type[Any]
) -> Optional[Callable[[Any], Any]]:
  """Get converter from source type to destination type."""
  return _TYPE_CONVERTER_REGISTRY.get_converter(src, dest)


def get_first_applicable_converter(
    src_type: Type[Any],
    dest_type: Union[Type[Any], Tuple[Type[Any], ...]]):
  """Get first applicable converter."""
  dest_types = dest_type if isinstance(dest_type, tuple) else (dest_type,)
  for dest_type in dest_types:
    converter = get_converter(src_type, dest_type)
    if converter is not None:
      return converter
  return None


def get_json_value_converter(src: Type[Any]) -> Optional[Callable[[Any], Any]]:
  """Get converter from source type to a JSON simple type."""
  return _TYPE_CONVERTER_REGISTRY.get_json_value_converter(src)


def register_converter(
    src_type: Union[Type[Any], Tuple[Type[Any], ...]],
    dest_type: Union[Type[Any], Tuple[Type[Any], ...]],
    convert_fn: Callable[[Any], Any]) -> None:
  """Register converter from source type to destination type.

  Examples::

    # Add converter from int to float.
    pg.typing.register_converter(int, float, float)

    assert pg.typing.Float().apply(1) is 1.0

    # Add converter from a dict to class A.
    def from_dict(d):
      return A(**d)

    assert isinstance(pg.typing.Object(A).apply({'x': 1, 'y': 2}), A)

  Args:
      src_type: Source value type.
      dest_type: Target value type.
      convert_fn: Function that performs the conversion, in signature
        (src_type) -> dest_type.
  """
  _TYPE_CONVERTER_REGISTRY.register(src_type, dest_type, convert_fn)


def _register_builtin_converters():
  """Register built-in converters."""
  # int <=> datetime.datetime.
  register_converter(int, datetime.datetime, datetime.datetime.utcfromtimestamp)
  register_converter(datetime.datetime, int,
                     lambda x: calendar.timegm(x.timetuple()))

  # string <=> KeyPath.
  register_converter(str, object_utils.KeyPath,
                     object_utils.KeyPath.parse)
  register_converter(object_utils.KeyPath, str, lambda x: x.path)


_register_builtin_converters()
object_utils.JSONConvertible.TYPE_CONVERTER = get_json_value_converter




# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for pyglove.hyper.DynamicEvaluationContext."""

import threading
import unittest

from pyglove.core import geno
from pyglove.core import symbolic
from pyglove.core.hyper.categorical import manyof
from pyglove.core.hyper.categorical import oneof
from pyglove.core.hyper.categorical import OneOf
from pyglove.core.hyper.categorical import permutate
from pyglove.core.hyper.custom import CustomHyper
from pyglove.core.hyper.dynamic_evaluation import dynamic_evaluate as pg_dynamic_evaluate
from pyglove.core.hyper.dynamic_evaluation import DynamicEvaluationContext
from pyglove.core.hyper.dynamic_evaluation import trace as pg_trace
from pyglove.core.hyper.numerical import floatv
from pyglove.core.hyper.object_template import template as pg_template


class DynamicEvaluationTest(unittest.TestCase):
  """Dynamic evaluation test."""

  def test_dynamic_evaluate(self):
    with self.assertRaisesRegex(
        ValueError, '\'evaluate_fn\' must be either None or a callable object'):
      with pg_dynamic_evaluate(1):
        pass

    with self.assertRaisesRegex(
        ValueError, '\'exit_fn\' must be a callable object'):
      with pg_dynamic_evaluate(None, exit_fn=1):
        pass

  def test_evaluated_values_during_collect(self):
    with DynamicEvaluationContext().collect():
      self.assertEqual(oneof([0, 1]), 0)
      self.assertEqual(oneof([{'x': oneof(['a', 'b'])}, 1]),
                       {'x': 'a'})
      self.assertEqual(manyof(2, [0, 1, 3]), [0, 1])
      self.assertEqual(manyof(4, [0, 1, 3], distinct=False),
                       [0, 0, 0, 0])
      self.assertEqual(permutate([0, 1, 2]), [0, 1, 2])
      self.assertEqual(floatv(0.0, 1.0), 0.0)

  def test_per_thread_collect_and_apply(self):
    def thread_fun():
      context = DynamicEvaluationContext()
      with context.collect():
        oneof(range(10))

      with context.apply([3]):
        self.assertEqual(oneof(range(10)), 3)

    threads = []
    for _ in range(10):
      thread = threading.Thread(target=thread_fun)
      threads.append(thread)
      thread.start()
    for t in threads:
      t.join()

  def test_process_wise_collect_and_apply(self):
    def thread_fun():
      _ = oneof(range(10))

    context = DynamicEvaluationContext(per_thread=False)
    with context.collect() as hyper_dict:
      threads = []
      for _ in range(10):
        thread = threading.Thread(target=thread_fun)
        threads.append(thread)
        thread.start()
      for t in threads:
        t.join()
    self.assertEqual(len(hyper_dict), 10)

  def test_search_space_defined_without_hyper_name(self):
    def fun():
      x = oneof([1, 2, 3]) + 1
      y = sum(manyof(2, [2, 4, 6, 8], name='y'))
      z = floatv(min_value=1.0, max_value=2.0)
      return x + y + z

    # Test dynamic evaluation by allowing reentry (all hyper primitives will
    # be registered twice).
    context = DynamicEvaluationContext()
    with context.collect() as hyper_dict:
      result = fun()
      result = fun()

    # 1 + 1 + 2 + 4 + 1.0
    self.assertEqual(result, 9.0)
    self.assertEqual(hyper_dict, {
        'decision_0': oneof([1, 2, 3]),
        'y': manyof(2, [2, 4, 6, 8], name='y'),
        'decision_1': floatv(min_value=1.0, max_value=2.0),
        'decision_2': oneof([1, 2, 3]),
        'decision_3': floatv(min_value=1.0, max_value=2.0),
    })

    with context.apply(geno.DNA.parse(
        [1, [0, 2], 1.5, 0, 1.8])):
      # 2 + 1 + 2 + 6 + 1.5
      self.assertEqual(fun(), 12.5)
      # 1 + 1 + 2 + 6 + 1.8
      self.assertEqual(fun(), 11.8)

  def test_search_space_defined_with_hyper_name(self):
    def fun():
      x = oneof([1, 2, 3], name='a') + 1
      y = sum(manyof(2, [2, 4, 6, 8], name='b'))
      z = floatv(min_value=1.0, max_value=2.0, name='c')
      return x + y + z

    # Test dynamic evaluation by disallowing reentry (all hyper primitives will
    # be registered only once).
    context = DynamicEvaluationContext(require_hyper_name=True)
    with context.collect() as hyper_dict:
      with self.assertRaisesRegex(
          ValueError, '\'name\' must be specified for hyper primitive'):
        oneof([1, 2, 3])
      result = fun()
      result = fun()

    # 1 + 1 + 2 + 4 + 1.0
    self.assertEqual(result, 9.0)
    self.assertEqual(hyper_dict, symbolic.Dict(
        a=oneof([1, 2, 3], name='a'),
        b=manyof(2, [2, 4, 6, 8], name='b'),
        c=floatv(min_value=1.0, max_value=2.0, name='c')))
    with context.apply(geno.DNA.parse([1, [0, 2], 1.5])):
      # We can call fun multiple times since decision will be bound to each
      # name just once.
      # 2 + 1 + 2 + 6 + 1.5
      self.assertEqual(fun(), 12.5)
      self.assertEqual(fun(), 12.5)
      self.assertEqual(fun(), 12.5)

  def test_hierarchical_search_space(self):
    def fun():
      return oneof([
          lambda: sum(manyof(2, [2, 4, 6, 8])),
          lambda: oneof([3, 7]),
          lambda: floatv(min_value=1.0, max_value=2.0),
          10]) + oneof([11, 22])

    context = DynamicEvaluationContext()
    with context.collect() as hyper_dict:
      result = fun()
    # 2 + 4 + 11
    self.assertEqual(result, 17)
    self.assertEqual(hyper_dict, {
        'decision_0': oneof([
            # NOTE(daiyip): child decisions within candidates are always in
            # form of list.
            {
                'decision_1': manyof(2, [2, 4, 6, 8]),
            },
            {
                'decision_2': oneof([3, 7])
            },
            {
                'decision_3': floatv(min_value=1.0, max_value=2.0)
            },
            10,
        ]),
        'decision_4': oneof([11, 22])
    })

    with context.apply(geno.DNA.parse([(0, [1, 3]), 0])):
      # 4 + 8 + 11
      self.assertEqual(fun(), 23)

    # Use list-form decisions.
    with context.apply([0, 1, 3, 0]):
      # 4 + 8 + 11
      self.assertEqual(fun(), 23)

    with context.apply(geno.DNA.parse([(1, 1), 1])):
      # 7 + 22
      self.assertEqual(fun(), 29)

    with context.apply(geno.DNA.parse([(2, 1.5), 0])):
      # 1.5 + 11
      self.assertEqual(fun(), 12.5)

    with context.apply(geno.DNA.parse([3, 1])):
      # 10 + 22
      self.assertEqual(fun(), 32)

    with self.assertRaisesRegex(
        ValueError, '`decisions` should be a DNA or a list of numbers.'):
      with context.apply(3):
        fun()

    with self.assertRaisesRegex(
        ValueError, 'No decision is provided for .*'):
      with context.apply(geno.DNA.parse(3)):
        fun()

    with self.assertRaisesRegex(
        ValueError, 'Expect float-type decision for .*'):
      with context.apply([2, 0, 1]):
        fun()

    with self.assertRaisesRegex(
        ValueError, 'Expect int-type decision in range .*'):
      with context.apply([5, 0.5, 0]):
        fun()

    with self.assertRaisesRegex(
        ValueError, 'Found extra decision values that are not used.*'):
      with context.apply(geno.DNA.parse([(1, 1), 1, 1])):
        fun()

  def test_hierarchical_search_space_with_hyper_name(self):
    def fun():
      return oneof([
          lambda: sum(manyof(2, [2, 4, 6, 8], name='a1')),
          lambda: oneof([3, 7], name='a2'),
          lambda: floatv(min_value=1.0, max_value=2.0, name='a3.xx'),
          10], name='a') + oneof([11, 22], name='b')

    context = DynamicEvaluationContext(require_hyper_name=True)
    with context.collect() as hyper_dict:
      result = fun()
      result = fun()

    # 2 + 4 + 11
    self.assertEqual(result, 17)
    self.assertEqual(hyper_dict, {
        'a': oneof([
            # NOTE(daiyip): child decisions within candidates are always in
            # form of list.
            {'a1': manyof(2, [2, 4, 6, 8], name='a1')},
            {'a2': oneof([3, 7], name='a2')},
            {'a3.xx': floatv(min_value=1.0, max_value=2.0, name='a3.xx')},
            10,
        ], name='a'),
        'b': oneof([11, 22], name='b')
    })

    with context.apply(geno.DNA.parse([(0, [1, 3]), 0])):
      # 4 + 8 + 11
      self.assertEqual(fun(), 23)
      self.assertEqual(fun(), 23)
      self.assertEqual(fun(), 23)

    # Use list form.
    with context.apply([0, 1, 3, 0]):
      # 4 + 8 + 11
      self.assertEqual(fun(), 23)
      self.assertEqual(fun(), 23)
      self.assertEqual(fun(), 23)

    with context.apply(geno.DNA.parse([(1, 1), 1])):
      # 7 + 22
      self.assertEqual(fun(), 29)
      self.assertEqual(fun(), 29)

    with context.apply(geno.DNA.parse([(2, 1.5), 0])):
      # 1.5 + 11
      self.assertEqual(fun(), 12.5)
      self.assertEqual(fun(), 12.5)

    with context.apply(geno.DNA.parse([3, 1])):
      # 10 + 22
      self.assertEqual(fun(), 32)
      self.assertEqual(fun(), 32)

    with self.assertRaisesRegex(
        ValueError, '`decisions` should be a DNA or a list of numbers.'):
      with context.apply(3):
        fun()

    with self.assertRaisesRegex(
        ValueError, 'DNA value type mismatch'):
      with context.apply(geno.DNA.parse(3)):
        fun()

    with self.assertRaisesRegex(
        ValueError, 'Found extra decision values that are not used'):
      with context.apply(context.dna_spec.first_dna()):
        # Do not consume any decision points from the search space.
        _ = 1

    with self.assertRaisesRegex(
        ValueError,
        'Hyper primitive .* is not defined during search space inspection'):
      with context.apply(context.dna_spec.first_dna()):
        # Do not consume any decision points from the search space.
        _ = oneof(range(5), name='uknown')

  def test_where_statement(self):
    context = DynamicEvaluationContext(
        where=lambda x: getattr(x, 'name') != 'x')
    with context.collect():
      self.assertEqual(oneof(range(10)), 0)
      self.assertIsInstance(oneof(range(5), name='x'), OneOf)

    with context.apply([1]):
      self.assertEqual(oneof(range(10)), 1)
      self.assertIsInstance(oneof(range(5), name='x'), OneOf)

  def test_trace(self):
    def fun():
      return oneof([-1, 0, 1]) * oneof([-1, 0, 3]) + 1

    self.assertEqual(
        pg_trace(fun).hyper_dict,
        {
            'decision_0': oneof([-1, 0, 1]),
            'decision_1': oneof([-1, 0, 3])
        })

  def test_dynamic_evaluation_with_custom_hyper(self):

    class IntList(CustomHyper):

      def custom_decode(self, dna):
        return [int(x) for x in dna.value.split(':')]

      def first_dna(self):
        return geno.DNA('0:1:2:3')

    def fun():
      return sum(IntList()) + oneof([0, 1]) + floatv(-1., 1.)

    context = DynamicEvaluationContext()
    with context.collect():
      fun()

    self.assertEqual(
        context.hyper_dict,
        {
            'decision_0': IntList(),
            'decision_1': oneof([0, 1]),
            'decision_2': floatv(-1., 1.)
        })
    with context.apply(geno.DNA(['1:2:3:4', 1, 0.5])):
      self.assertEqual(fun(), 1 + 2 + 3 + 4 + 1 + 0.5)

    with self.assertRaisesRegex(
        ValueError, 'Expect string-type decision for .*'):
      with context.apply(geno.DNA([0, 1, 0.5])):
        fun()

    class IntListWithoutFirstDNA(CustomHyper):

      def custom_decode(self, dna):
        return [int(x) for x in dna.value.split(':')]

    context = DynamicEvaluationContext()
    with self.assertRaisesRegex(
        NotImplementedError,
        '.* must implement method `next_dna` to be used in '
        'dynamic evaluation mode'):
      with context.collect():
        IntListWithoutFirstDNA()

  def test_dynamic_evaluation_with_external_dna_spec(self):
    def fun():
      return oneof(range(5), name='x') + oneof(range(3), name='y')

    context = pg_trace(fun, require_hyper_name=True, per_thread=True)
    self.assertFalse(context.is_external)
    self.assertIsNotNone(context.hyper_dict)

    search_space_str = symbolic.to_json_str(context.dna_spec)

    context2 = DynamicEvaluationContext(
        require_hyper_name=True, per_thread=True,
        dna_spec=symbolic.from_json_str(search_space_str))
    self.assertTrue(context2.is_external)
    self.assertIsNone(context2.hyper_dict)

    with self.assertRaisesRegex(
        ValueError,
        '`collect` cannot be called .* is using an external DNASpec'):
      with context2.collect():
        fun()

    with context2.apply(geno.DNA([1, 2])):
      self.assertEqual(fun(), 3)

  def test_search_space_partitioning_without_hyper_name(self):
    def fun():
      return sum([
          oneof([1, 2, 3], hints='ssd1'),
          oneof([4, 5], hints='ssd2'),
      ])

    context1 = DynamicEvaluationContext(where=lambda x: x.hints == 'ssd1')
    context2 = DynamicEvaluationContext(where=lambda x: x.hints == 'ssd2')
    with context1.collect():
      with context2.collect():
        self.assertEqual(fun(), 1 + 4)

    self.assertEqual(
        context1.hyper_dict, {
            'decision_0': oneof([1, 2, 3], hints='ssd1')
        })
    self.assertEqual(
        context2.hyper_dict, {
            'decision_0': oneof([4, 5], hints='ssd2')
        })
    with context1.apply(geno.DNA(2)):
      with context2.apply(geno.DNA(1)):
        self.assertEqual(fun(), 3 + 5)

  def test_search_space_partitioning_with_hyper_name(self):
    def fun():
      return sum([
          oneof([1, 2, 3], name='x', hints='ssd1'),
          oneof([4, 5], name='y', hints='ssd2'),
      ])

    context1 = DynamicEvaluationContext(where=lambda x: x.hints == 'ssd1')
    context2 = DynamicEvaluationContext(where=lambda x: x.hints == 'ssd2')
    with context1.collect():
      with context2.collect():
        self.assertEqual(fun(), 1 + 4)

    self.assertEqual(
        context1.hyper_dict, {
            'x': oneof([1, 2, 3], name='x', hints='ssd1')
        })
    self.assertEqual(
        context2.hyper_dict, {
            'y': oneof([4, 5], name='y', hints='ssd2')
        })
    with context1.apply(geno.DNA(2)):
      with context2.apply(geno.DNA(1)):
        self.assertEqual(fun(), 3 + 5)

  def test_hierarchial_search_space_with_partitioning(self):
    def fun():
      return sum([
          oneof([
              lambda: oneof([1, 2, 3], name='y', hints='ssd1'),
              lambda: oneof([4, 5, 6], name='z', hints='ssd1'),
          ], name='x', hints='ssd1'),
          oneof([7, 8], name='p', hints='ssd2'),
          oneof([9, 10], name='q', hints='ssd2'),
      ])
    context1 = DynamicEvaluationContext(where=lambda x: x.hints == 'ssd1')
    context2 = DynamicEvaluationContext(where=lambda x: x.hints == 'ssd2')
    with context1.collect():
      with context2.collect():
        self.assertEqual(fun(), 1 + 7 + 9)

    self.assertEqual(
        context1.hyper_dict, {
            'x': oneof([
                {'y': oneof([1, 2, 3], name='y', hints='ssd1')},
                {'z': oneof([4, 5, 6], name='z', hints='ssd1')},
            ], name='x', hints='ssd1')
        })
    self.assertEqual(
        context2.hyper_dict, {
            'p': oneof([7, 8], name='p', hints='ssd2'),
            'q': oneof([9, 10], name='q', hints='ssd2')
        })
    with context1.apply(geno.DNA((1, 1))):
      with context2.apply(geno.DNA([0, 1])):
        self.assertEqual(fun(), 5 + 7 + 10)

  def test_search_space_partitioning_with_different_per_thread_settings(self):
    context1 = DynamicEvaluationContext(per_thread=True)
    context2 = DynamicEvaluationContext(per_thread=False)

    def fun():
      return oneof([1, 2, 3])

    with self.assertRaisesRegex(
        ValueError,
        'Nested dynamic evaluation contexts must be either .*'):
      with context1.collect():
        with context2.collect():
          fun()

  def test_manual_decision_point_registration(self):
    context = DynamicEvaluationContext()
    self.assertEqual(
        context.add_decision_point(oneof([1, 2, 3])), 1)
    self.assertEqual(
        context.add_decision_point(oneof(['a', 'b'], name='x')), 'a')
    self.assertEqual(
        context.add_decision_point(pg_template(1)), 1)

    with self.assertRaisesRegex(
        ValueError, 'Found different hyper primitives under the same name'):
      context.add_decision_point(oneof(['foo', 'bar'], name='x'))

    self.assertEqual(context.hyper_dict, {
        'decision_0': oneof([1, 2, 3]),
        'x': oneof(['a', 'b'], name='x'),
    })

    with self.assertRaisesRegex(
        ValueError, '`evaluate` needs to be called under the `apply` context'):
      context.evaluate(oneof([1, 2, 3]))

    with context.apply([1, 1]):
      self.assertEqual(context.evaluate(context.hyper_dict['decision_0']), 2)
      self.assertEqual(context.evaluate(context.hyper_dict['x']), 'b')


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for pyglove.symbolic.ContextualGetter."""

import unittest

from pyglove.core.symbolic import base
from pyglove.core.symbolic import contextual
from pyglove.core.symbolic.dict import Dict


class ContextualGetterTest(unittest.TestCase):
  """Tests for `pg.symbolic.ContextualGetter`."""

  def test_basics(self):
    @contextual.contextual_getter
    def static_value(context, v):
      del context
      return v

    getter = static_value(v=1)  # pylint: disable=no-value-for-parameter
    self.assertIsInstance(getter, base.ContextualValue)
    self.assertEqual(
        getter.get(base.GetAttributeContext('x', Dict(), Dict())), 1
    )
    self.assertEqual(base.from_json(base.to_json(getter)), getter)


if __name__ == '__main__':
  unittest.main()
